

class NotSupportedError(Exception):

    def __init__(self, message: str) -> None:
        super().__init__(message)


class HeaderTypeDescriptorError(Exception):

    def __init__(self, message: str) -> None:
        super().__init__(message)
